#include <udm_config.h>

#include <stdio.h>
#include <stdlib.h>

#ifdef HAVE_ZLIB
#include <zlib.h>

#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/errno.h>
#include <unistd.h>
#include <signal.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <fcntl.h>
#include <netdb.h>
#ifdef HAVE_SYS_WAIT_H
#include <sys/wait.h>
#endif

#include "udm_common.h"
#include "udm_utils.h"
#include "udm_env.h"
#include "udm_conf.h"
#include "udm_services.h"
#include "udm_store.h"

#define UDM_STORED_MAXCLNTS 128

static char pidname[1024];
static char time_pid[100];
static UDM_ENV * Conf=NULL;


static void exitproc(void){
	unlink(pidname);
}

static char * time_pid_info(void){
	struct tm * tim;
	time_t t;
	t=time(NULL);
	tim=localtime(&t);
	strftime(time_pid,sizeof(time_pid),"%a %d %T",tim);
	sprintf(time_pid+strlen(time_pid)," [%d]",(int)getpid());
	return(time_pid);
}

/********************* SIG Handlers ****************/
static void sighandler(int sign);
static void init_signals(void){
	/* Set up signals handler*/
	signal(SIGPIPE,sighandler);
	signal(SIGHUP,sighandler);
	signal(SIGINT,sighandler);
	signal(SIGTERM,sighandler);
	signal(SIGCHLD, sighandler);
}
static void sighandler(int sign){
  int status;

	fprintf(stderr,"%s Signal %d arrived\n",time_pid_info(),sign);
	switch(sign){
		case SIGPIPE:
			fprintf(stderr,"%s FATAL: This should not ever happen! Broken pipe!\n",time_pid_info());
			break;
		case SIGHUP:
			break;
	        case SIGCHLD:
		  wait(&status);
		  break;
		default:
			fprintf(stderr,"%s Shutdown\n",time_pid_info());
			exit(0);
	}
	init_signals();
}

ssize_t recvall(int s, void *buf, size_t len, int flags);

ssize_t recvall(int s, void *buf, size_t len, int flags) {
  size_t received = 0;
  char *b = buf;
  while ((received += recv(s, &b[received], len - received, flags)) < len);
  return received;
}

#define UDM_STOREIND_SIG "msINDEX\0"
#define UDM_STORE_SIG "msSTORE\0"
#define UDM_SIG_LEN 8
#define UDM_STORE_BITS 20      /* number bits of rec_id per page */
#define UDM_STORE_MASK ((1L << UDM_STORE_BITS) - 1)

typedef struct StoreIndexHeader {
  const char sig[UDM_SIG_LEN];
  size_t nrec;
  size_t nsorted;
} UDM_STOREIND_HEADER;

typedef struct StoreIndexItem {
  int rec_id;
  off_t offset;
  size_t size;
} UDM_STOREIND_ITEM;


#endif

int main(int argc, char **argv) {

#ifdef HAVE_ZLIB

	struct sockaddr_in server_addr, client_addr;
	const char *config_name= UDM_CONF_DIR "/indexer.conf";
	const char *store_dir = UDM_VAR_DIR "/store/";
	int pid_fd, nport, s, on = 1;
	char pidbuf[1024];
	char buf[1024];

	/* Check that another instance isn't running */
	/* and create PID file.                      */

	sprintf(pidname,"%s/%s",UDM_VAR_DIR,"stored.pid");
	pid_fd = open(pidname,O_CREAT|O_EXCL|O_WRONLY,0644);
	if(pid_fd < 0){
		fprintf(stderr,"%s Can't create '%s': %s\n", time_pid_info(), pidname, strerror(errno));
		if(errno == EEXIST){
			fprintf(stderr,"It seems that another urld is already running!\n");
			fprintf(stderr,"Remove '%s' if it is not true.\n",pidname);
 		}
		exit(1);
	}

	sprintf(pidbuf,"%d\n",(int)getpid());
	write(pid_fd,&pidbuf,strlen(pidbuf));

	atexit(&exitproc);
	init_signals();
	UdmInit(); /* Initialize library */

	Conf = UdmAllocEnv();

	UdmLoadConfig(Conf,config_name,0,0);

	if(UdmEnvErrCode(Conf)){
 		fprintf(stderr,"%s\n",UdmEnvErrMsg(Conf));
		UdmFreeEnv(Conf);
		unlink(pidname);
		exit(1);
	}

	nport = UDM_STORED_PORT;
	nport = htons((u_short) nport);

	if ((s = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
		fprintf(stderr,"%s socket() error %d\n",time_pid_info(),errno);
		UdmFreeEnv(Conf);
		unlink(pidname);
		exit(1);
	}
	if (setsockopt(s, SOL_SOCKET, SO_REUSEADDR, (char *)&on, sizeof(on)) != 0){
		fprintf(stderr,"%s setsockopt() error %d\n",time_pid_info(),errno);
		UdmFreeEnv(Conf);
		unlink(pidname);
		exit(1);
	}

	bzero(&server_addr, sizeof(server_addr));
	server_addr.sin_family = AF_INET;
	server_addr.sin_addr.s_addr = htonl(INADDR_ANY);
	server_addr.sin_port = nport;
	

	if (bind(s, (struct sockaddr *)&server_addr, sizeof(server_addr)) == -1) {
		fprintf(stderr,"%s bind() error %d %s\n",time_pid_info(),errno,strerror(errno));
		unlink(pidname);
		UdmFreeEnv(Conf);
		exit(1);
	}

	if (listen(s, UDM_STORED_MAXCLNTS) == -1) {
		fprintf(stderr,"%s listen() error %d %s\n",time_pid_info(),errno,strerror(errno));
		unlink(pidname);
		UdmFreeEnv(Conf);
		exit(1);
	}
	
	while(1) {
	  size_t addrlen;
	  int ns, pid;

	  bzero(&client_addr, addrlen = sizeof(client_addr));

	  if ((ns = accept(s, (struct sockaddr *) &client_addr, &addrlen)) == -1) {
	    fprintf(stderr,"%s accept() error %d %s\n",time_pid_info(),errno,strerror(errno));
	    unlink(pidname);
	    exit(1);
	  }

	  fprintf(stderr, "Client = %s\n", inet_ntoa(client_addr.sin_addr));
    
	  if ((pid = fork() ) == -1) {
	    fprintf(stderr,"%s fork() error %d %s\n",time_pid_info(),errno,strerror(errno));
	    unlink(pidname);
	    exit(1);
	  }

	

	  if (pid == 0) { /* child process */
	    int rec_id;
	    size_t DocSize;
	    ssize_t nbytes;
	    Byte *Doc, *CDoc;
	    z_stream zstream;
	    

	    close(s);
      
	    if ((nbytes = recv(ns, buf, sizeof(buf), 0))) {
	      if (strncmp(buf, "STORE", 5) == 0) { 
		
		if (recvall(ns, &rec_id, sizeof(rec_id), MSG_WAITALL) < 0) {
		  closesocket(ns);
		  exit(1);
		}
		if (recvall(ns, &DocSize, sizeof(DocSize), MSG_WAITALL) < 0) {
		  closesocket(ns);
		  exit(1);
		}
		Doc = (Byte *) malloc(DocSize);
		if (Doc == NULL) {
		  closesocket(ns);
		  exit(1);
		}
		if (recvall(ns, Doc, DocSize, MSG_WAITALL) < 0) {
		  closesocket(ns);
		  exit(1);
		}

		zstream.zalloc = Z_NULL;
		zstream.zfree = Z_NULL;
		zstream.opaque = Z_NULL;
		zstream.next_in = Doc;
		
		if (deflateInit2(&zstream, 9, Z_DEFLATED, 15, 9, Z_DEFAULT_STRATEGY) == Z_OK) {
		
		  zstream.avail_in = DocSize;
		  CDoc = zstream.next_out = (Byte *) malloc(DocSize);
		  if (zstream.next_out == NULL) {
		    closesocket(ns);
		    exit(1);
		  }
		  deflate(&zstream, Z_FINISH);
		  deflateEnd(&zstream);

		  


		  UDM_FREE(zstream.next_out);
		}

		UDM_FREE(Doc);
		send (ns, "STORED", 6, 0);
		

	      } else {
		fprintf(stderr,"%s <hello> error %d %s\n",time_pid_info(),errno,buf /*strerror(errno)*/);
		close(ns);
		exit(1);
	      }
	    } else {
	      fprintf(stderr,"%s recv() <hello> error %d %s\n",time_pid_info(),errno,strerror(errno));
	      close(ns);
	      exit(1);
	    }


	    close(ns);
	    exit(0);
	  }
	  /* parent process */
	  close(ns);
	}

#else
	fprintf(stderr, "zlib support required. Please rebuild with --with-zlib option\n");

#endif
	return 0;
}
